/**
 * 20241029 ©WANGKANG 1686617586@qq.com
 */

package com.taais.biz.service.impl;

import cn.hutool.core.lang.Dict;
import cn.hutool.core.util.ObjectUtil;
import com.alibaba.fastjson2.JSON;
import com.mybatisflex.core.paginate.Page;
import com.mybatisflex.core.query.QueryWrapper;
import com.taais.biz.constant.BizConstant;
import com.taais.biz.domain.*;
import com.taais.biz.domain.bo.TrackSequenceBo;
import com.taais.biz.domain.vo.AlgorithmConfigTrackVo;
import com.taais.biz.domain.vo.AlgorithmModelTrackVo;
import com.taais.biz.domain.vo.StartTaskConfig;
import com.taais.biz.domain.vo.TrackSequenceVo;
import com.taais.biz.mapper.TrackSequenceMapper;
import com.taais.biz.service.ITrackSequenceService;
import com.taais.biz.utils.ZipUtils;
import com.taais.common.core.config.TaaisConfig;
import com.taais.common.core.constant.Constants;
import com.taais.common.core.core.domain.CommonResult;
import com.taais.common.core.core.page.PageResult;
import com.taais.common.core.utils.MapstructUtils;
import com.taais.common.core.utils.StringUtils;
import com.taais.common.json.utils.JsonUtils;
import com.taais.common.orm.core.page.PageQuery;
import com.taais.common.orm.core.service.impl.BaseServiceImpl;
import com.taais.common.websocket.utils.WebSocketUtils;
import com.taais.system.domain.SysOss;
import com.taais.system.domain.vo.SysOssVo;
import com.taais.system.service.ISysOssService;
import com.taais.system.service.impl.SysOssServiceImpl;

import jakarta.annotation.Resource;

import org.apache.commons.lang3.ObjectUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.FileSystemResource;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.DecimalFormat;
import java.util.*;

import static com.taais.biz.constant.BizConstant.VideoStatus.NOT_START;
import static com.taais.biz.domain.table.TrackSequenceTableDef.TRACK_SEQUENCE;
import static com.taais.biz.service.impl.TargetDetectionServiceImpl.*;
import static com.taais.biz.service.impl.ToInfraredServiceImpl.*;
import static com.taais.biz.service.impl.VideoStableServiceImpl.*;

/**
 * 注视轨迹序列Service业务层处理
 *
 * @author wangkang
 *         2024-09-22
 */
@Service
public class TrackSequenceServiceImpl extends BaseServiceImpl<TrackSequenceMapper, TrackSequence>
        implements ITrackSequenceService {

    private static final String MASC = "MASC";
    private static final String CAT = "CAT";

    @Autowired
    private AlgorithmConfigTrackServiceImpl algorithmConfigTrackService;
    @Autowired
    private AlgorithmModelTrackServiceImpl algorithmModelTrackService;

    @Autowired
    private ISysOssService ossService;

    @Resource
    private TrackSequenceMapper trackSequenceMapper;

    public List<TrackSequence> getByParentTaskId(Long parentTaskId) {
        QueryWrapper queryWrapper = super.buildBaseQueryWrapper();
        queryWrapper.and(TRACK_SEQUENCE.PARENT_TASK_ID.eq(parentTaskId));
        return this.list(queryWrapper);
    }

    // public String getTrainInputPath(SysOssVo ossEntity) {
    // return getUnZipDirPath(ossEntity);
    // }

    public String getTrainOutputPath(TrackSequence entity, SysOssVo ossEntity) {
        String resourcePath = getResourcePath(ossEntity);
        Path path = Paths.get(resourcePath);
        Path outputPath = path.resolveSibling(entity.getId().toString() + BizConstant.TRACK_SEQUENCE_SUFFIX);
        return outputPath.toString();
    }

    public String getPredictInputPath(SysOssVo ossEntity) {
        // todo
        return getUnZipDirPath(ossEntity);
    }

    public String getPredictOutputPath(TrackSequence entity, SysOssVo ossEntity) {
        return getTrainOutputPath(entity, ossEntity);
    }

    public String getTestInputPath(SysOssVo ossEntity) {
        // todo
        return getUnZipDirPath(ossEntity);
    }

    public String getTestOutputPath(TrackSequence entity, SysOssVo ossEntity) {
        // todo
        return getTrainOutputPath(entity, ossEntity);
    }

    private String getModelPath(Long modelId) {
        if (ObjectUtil.isNull(modelId)) {
            System.out.println("模型id为空");
            return "";
        }
        AlgorithmModelTrack algorithmModelTrack = algorithmModelTrackService.getById(modelId);
        if (ObjectUtil.isNull(algorithmModelTrack)) {
            System.out.println("模型不存在");
            return "";
        }
        return algorithmModelTrack.getModelAddress();
    }

    @Override
    public QueryWrapper query() {
        return super.query().from(TRACK_SEQUENCE);
    }

    private QueryWrapper buildQueryWrapper(TrackSequenceBo trackSequenceBo) {
        QueryWrapper queryWrapper = super.buildBaseQueryWrapper();
        queryWrapper.and(TRACK_SEQUENCE.NAME.like(trackSequenceBo.getName()));
        queryWrapper.and(TRACK_SEQUENCE.STATUS.eq(trackSequenceBo.getStatus()));
        queryWrapper.and(TRACK_SEQUENCE.START_TIME.eq(trackSequenceBo.getStartTime()));
        queryWrapper.and(TRACK_SEQUENCE.END_TIME.eq(trackSequenceBo.getEndTime()));
        queryWrapper.and(TRACK_SEQUENCE.COST_SECOND.eq(trackSequenceBo.getCostSecond()));
        queryWrapper.and(TRACK_SEQUENCE.LOG.eq(trackSequenceBo.getLog()));
        queryWrapper.and(TRACK_SEQUENCE.REMARKS.eq(trackSequenceBo.getRemarks()));
        queryWrapper.and(TRACK_SEQUENCE.URL.eq(trackSequenceBo.getUrl()));
        queryWrapper.and(TRACK_SEQUENCE.INPUT_OSS_ID.eq(trackSequenceBo.getInputOssId()));
        queryWrapper.and(TRACK_SEQUENCE.INPUT_PATH.eq(trackSequenceBo.getInputPath()));
        queryWrapper.and(TRACK_SEQUENCE.OUTPUT_PATH.eq(trackSequenceBo.getOutputPath()));
        queryWrapper.and(TRACK_SEQUENCE.ZIP_FILE_PATH.eq(trackSequenceBo.getZipFilePath()));

        return queryWrapper;
    }

    /**
     * 查询注视轨迹序列
     *
     * @param id 注视轨迹序列主键
     * @return 注视轨迹序列
     */
    @Override
    public TrackSequenceVo selectById(Long id) {
        return this.getOneAs(query().where(TRACK_SEQUENCE.ID.eq(id)), TrackSequenceVo.class);

    }

    /**
     * 查询注视轨迹序列列表
     *
     * @param trackSequenceBo 注视轨迹序列Bo
     * @return 注视轨迹序列集合
     */
    @Override
    public List<TrackSequenceVo> selectList(TrackSequenceBo trackSequenceBo) {
        QueryWrapper queryWrapper = buildQueryWrapper(trackSequenceBo);
        queryWrapper.orderBy(TRACK_SEQUENCE.CREATE_TIME, Boolean.FALSE);
        return this.listAs(queryWrapper, TrackSequenceVo.class);
    }

    /**
     * 分页查询注视轨迹序列列表
     *
     * @param trackSequenceBo 注视轨迹序列Bo
     * @return 分页注视轨迹序列集合
     */
    @Override
    public PageResult<TrackSequenceVo> selectPage(TrackSequenceBo trackSequenceBo) {
        QueryWrapper queryWrapper = buildQueryWrapper(trackSequenceBo);
        queryWrapper.orderBy(TRACK_SEQUENCE.CREATE_TIME, Boolean.FALSE);
        Page<TrackSequenceVo> page = this.pageAs(PageQuery.build(), queryWrapper, TrackSequenceVo.class);
        page.getRecords().forEach(entity -> {
            Long modelId = entity.getAlgorithmModelId();
            if (ObjectUtil.isNotNull(modelId)) {
                AlgorithmModelTrack model = algorithmModelTrackService.getById(modelId);
                entity.setModelName(model.getModelName());

            }

            Long algorithmId = entity.getAlgorithmId();
            if (ObjectUtil.isNotNull(algorithmId)) {
                AlgorithmConfigTrackVo config = algorithmConfigTrackService.selectById(algorithmId);
                entity.setType(config.getType());
                entity.setSubsystem(config.getSubsystem());
                entity.setAlgorithmName(config.getAlgorithmName());
            }
        });
        page.getRecords().sort(Comparator.comparing(TrackSequenceVo::getCreateTime).reversed());
        return PageResult.build(page);
    }

    public TrackSequence packageTrackSequence(TrackSequenceBo entityBo, SysOssVo ossEntity) {
        TrackSequence entity = new TrackSequence();
        entity.setInputOssId(entityBo.getInputOssId());
        entity.setUrl(ossEntity.getUrl());
        entity.setZipFilePath(ossEntity.getFileName());
        entity.setName(entityBo.getName());
        entity.setStatus(NOT_START);
        entity.setRemarks(entityBo.getRemarks());
        entity.setAlgorithmModelId(entityBo.getAlgorithmModelId());
        entity.setAlgorithmId(entityBo.getAlgorithmId());
        entity.setParentTaskId(entityBo.getParentTaskId());
        entity.setAlgorithmModelTargetDetectionId(entityBo.getAlgorithmModelTargetDetectionId());
        entity.setInputLabelOssId(entityBo.getInputLabelOssId());
        entity.setPredictTaskId(entityBo.getPredictTaskId());
        return entity;
    }

    /**
     * 新增注视轨迹序列
     *
     * @param entityBo 注视轨迹序列Bo
     * @return 结果:true 操作成功，false 操作失败
     */
    @Override
    @Transactional
    public CommonResult insert(TrackSequenceBo entityBo) {
        Long ossId = entityBo.getInputOssId();
        if (ObjectUtil.isNotEmpty(entityBo.getPredictTaskId())) { // 如果存在预测任务id，则使用label ossId
            ossId = entityBo.getInputLabelOssId();
        }

        // 检查input_oss_id是否存在
        if (ObjectUtil.isNull(ossId)) {
            throw new RuntimeException("oss文件不存在");
        }

        SysOssVo ossEntity = ossService.getById(ossId);
        if (ObjectUtil.isNull(ossEntity)) {
            throw new RuntimeException("oss文件不存在");
        }

        TrackSequence entity = packageTrackSequence(entityBo, ossEntity);
        boolean flag = this.save(entity);

        if (!flag) {
            throw new RuntimeException("新增注释轨迹序列任务失败");
        }

        entity = updateEntity(entity, entityBo, ossEntity);

        // 步骤 6. 保存算法参数到数据库
        boolean __ = this.updateById(entity);// 使用全局配置的雪花算法主键生成器生成ID值
        if (!__) {
            throw new RuntimeException("新增注释轨迹序列任务失败");
        }
        return CommonResult.success(entity);
    }

    private TrackSequence updateEntity(TrackSequence entity, TrackSequenceBo entityBo, SysOssVo ossEntity) {
        // 从这里开始，配置任务的algorithm_parameters参数
        // 步骤 1. 首先根据算法id获取算法配置
        AlgorithmConfigTrack algorithmConfig = algorithmConfigTrackService.getById(entity.getAlgorithmId());
        if (ObjectUtil.isNull(algorithmConfig)) {
            throw new RuntimeException("算法配置参数为空");
        }

        Map<String, Object> algorithmParameters = getAlgorithmParameters(algorithmConfig.getParameters(),
                entityBo.getOtherParams());

        // 步骤4. 构造可以直接传给前端的map数据结构
        Map<String, Object> result = new HashMap<>();

        result.put("bizId", entity.getId());
        result.put("bizType", BizConstant.BizType.TRACK_SEQUENCE);

        if (algorithmConfig.getType().equals(BizConstant.AlgorithmType.TRAIN)) {
            String source_dir = getTrainInputPath(ossEntity);
            String result_dir = getTrainOutputPath(entity, ossEntity);
            String log_path = getLogFilePath(result_dir, entity.getId(), BizConstant.TRACK_SEQUENCE_SUFFIX);
            result.put("source_dir", source_dir);
            result.put("result_dir", result_dir);
            result.put("log_path", log_path);
        } else if (algorithmConfig.getType().equals(BizConstant.AlgorithmType.REASONING)) {
            String source_dir = getPredictInputPath(ossEntity);
            String result_dir = getPredictOutputPath(entity, ossEntity);
            String log_path = getLogFilePath(result_dir, entity.getId(), BizConstant.TRACK_SEQUENCE_SUFFIX);

            result.put("source_dir", source_dir);
            result.put("result_dir", result_dir);
            result.put("log_path", log_path);

            if (ObjectUtil.isNotEmpty(entity.getAlgorithmModelId())) {
                String model_path = getModelPath(entity.getAlgorithmModelId());
                result.put("model_path", model_path);
            }

            if (ObjectUtil.isNotEmpty(entity.getAlgorithmModelTargetDetectionId())) {
                String model_path_TD = getModelPath(entity.getAlgorithmModelTargetDetectionId());
                result.put("model_path_TD", model_path_TD);
            } else {
                throw new RuntimeException("目标检测模型不存在!");
            }
        } else if (algorithmConfig.getType().equals(BizConstant.AlgorithmType.TEST)) {
            if (ObjectUtil.isEmpty(entityBo.getPredictTaskId())) {
                throw new RuntimeException("预测任务id为空");
            }
            TrackSequence predictTask = getById(entityBo.getPredictTaskId());
            if (ObjectUtil.isNull(predictTask)) {
                throw new RuntimeException("预测任务不存在");
            }

            String source_dir = predictTask.getInputPath();
            String txt_dir = predictTask.getOutputPath();
            String result_dir = getTestOutputPath(entity, ossEntity);
            String log_path = getLogFilePath(result_dir, entity.getId(), BizConstant.TRACK_SEQUENCE_SUFFIX);
            String label_dir = getResourcePath(ossEntity);
            result.put("source_dir", source_dir);
            result.put("result_dir", result_dir);
            result.put("log_path", log_path);
            result.put("label_dir", label_dir);
            result.put("txt_dir", txt_dir);
        } else {
            throw new RuntimeException("算法类型错误");
        }

        result.put("otherParams", algorithmParameters);

        entity.setInputPath((String) result.get("source_dir"));
        entity.setOutputPath((String) result.get("result_dir"));
        entity.setLogPath((String) result.get("log_path"));
        // 步骤 5. 将算法参数map序列化为json字符串，保存到数据库中
        entity.setAlgorithmParameters(JsonUtils.toJsonString(result));

        makeDir(entity.getInputPath());
        makeDir(entity.getOutputPath());

        File file = new File(entity.getInputPath());
        if (!file.exists()) {
            String resourcePath = getResourcePath(ossService.getById(entity.getInputOssId()));
            ZipUtils.unzip(resourcePath, entity.getInputPath());
        }

        return entity;
    }

    /**
     * 修改注视轨迹序列
     *
     * @param trackSequenceBo 注视轨迹序列Bo
     * @return 结果:true 更新成功，false 更新失败
     */
    @Override
    public boolean update(TrackSequenceBo trackSequenceBo) {
        TrackSequence entity = MapstructUtils.convert(trackSequenceBo, TrackSequence.class);

        Long ossId = trackSequenceBo.getInputOssId();
        if (ObjectUtil.isNotEmpty(trackSequenceBo.getPredictTaskId())) { // 如果存在预测任务id，则使用label ossId
            ossId = trackSequenceBo.getInputLabelOssId();
        }

        // 检查input_oss_id是否存在
        if (ObjectUtil.isNull(ossId)) {
            throw new RuntimeException("oss文件不存在");
        }

        SysOssVo ossEntity = ossService.getById(ossId);
        if (ObjectUtil.isNull(ossEntity)) {
            throw new RuntimeException("oss文件不存在");
        }

        entity = updateEntity(entity, trackSequenceBo, ossEntity);

        if (ObjectUtil.isNotNull(entity) && ObjectUtil.isNotNull(entity.getId())) {
            boolean updated = this.updateById(entity);
            return updated;
        }
        return false;
    }

    /**
     * 批量删除注视轨迹序列
     *
     * @param ids 需要删除的注视轨迹序列主键集合
     * @return 结果:true 删除成功，false 删除失败
     */
    @Transactional
    @Override
    public boolean deleteByIds(Long[] ids) {
        return this.removeByIds(Arrays.asList(ids));
    }

    @Override
    public CommonResult start(Long id) {
        TrackSequence entity = getById(id);
        entity.setStartTime(new Date());

        AlgorithmConfigTrack algorithmConfigTrack = algorithmConfigTrackService.getById(entity.getAlgorithmId());
        Map<String, Object> startTaskConfig = JsonUtils.parseMap(entity.getAlgorithmParameters());

        HttpResponseEntity responseEntity = sendPostMsg(algorithmConfigTrack.getStartApi(), startTaskConfig);
        if (ObjectUtil.isNotNull(responseEntity) && responseEntity.getStatus() == 200) {
            entity.setStatus(BizConstant.VideoStatus.RUNNING);
            updateById(entity);
            return CommonResult.success("任务开始成功，请等待完成");
        } else {
            entity.setStatus(BizConstant.VideoStatus.FAILED);
            updateById(entity);
            return CommonResult.fail("任务开始失败，请检查！");
        }
    }

    @Override
    public CommonResult stop(Long id) {
        TrackSequence entity = getById(id);

        StartTaskConfig startTaskConfig = new StartTaskConfig();
        startTaskConfig.setBizType(BizConstant.BizType.TRACK_SEQUENCE);
        startTaskConfig.setBizId(entity.getId());

        AlgorithmConfigTrack algorithmConfigTrack = algorithmConfigTrackService.getById(entity.getAlgorithmId());

        HttpResponseEntity responseEntity = sendPostMsg(algorithmConfigTrack.getTerminateApi(), startTaskConfig);
        if (ObjectUtil.isNotNull(responseEntity) && responseEntity.getStatus() == 200) {
            entity.setStatus(BizConstant.VideoStatus.INTERRUPTED);
            updateById(entity);
            return CommonResult.success("终止任务成功");
        } else {
            return CommonResult.fail("终止任务失败");
        }
    }

    @Override
    public ResponseEntity<org.springframework.core.io.Resource> zipImages(Long id) {
        TrackSequence entity = this.getById(id);
        if (ObjectUtil.isNull(entity)) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(null);
        }

        String outputPath = entity.getOutputPath();
        String zipFilePath = outputPath + ".zip";

        File file = new File(zipFilePath);
        if (!file.exists()) {
            try {
                ZipUtils.zipFolderFiles(outputPath, zipFilePath);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        if (!file.exists() || !file.isFile()) {
            return ResponseEntity.status(HttpStatus.NOT_FOUND).body(null);
        }

        org.springframework.core.io.Resource resource = new FileSystemResource(file);
        return ResponseEntity.ok()
                .header(HttpHeaders.CONTENT_DISPOSITION, "attachment; filename=\"" + file.getName() + "\"")
                .header(HttpHeaders.CONTENT_TYPE, "application/octet-stream").body(resource);
    }

    @Override
    public CommonResult getResult(TaskTrackResultBo taskTrackResultBo) {
        Long id = taskTrackResultBo.getBizId();
        String status = taskTrackResultBo.getStatus();
        String msg = taskTrackResultBo.getMsg();
        TrackSequence entity = getById(id);
        entity.setLog(msg);
        entity.setStatus("200".equals(status) ? BizConstant.VideoStatus.END : BizConstant.VideoStatus.FAILED);
        entity.setEndTime(new Date());
        try {
            entity.setCostSecond((entity.getEndTime().getTime() - entity.getStartTime().getTime()) / 1000);
        } catch (Exception e) {
            entity.setCostSecond(null);
        }
        updateById(entity);

        WebSocketUtils.publishAll("refresh");
        return CommonResult.success();
    }

    @Override
    public CommonResult getLog(Long id) {
        TrackSequence entity = getById(id);
        File file = new File(entity.getLogPath());
        if (!file.exists()) {
            return CommonResult.fail("日志文件不存在！");
        }
        return CommonResult.success(readLogContent(entity.getLogPath()), "success");
    }

    @Override
    public CommonResult getDetail(Long id) {
        TrackSequenceVo entity = selectById(id);
        if (entity.getName().startsWith(MASC)) {
            AlgorithmModelTrack modelTrack = algorithmModelTrackService.getById(entity.getAlgorithmModelId());
            AlgorithmConfigTrack algorithmConfigTrack = algorithmConfigTrackService.getById(entity.getAlgorithmId());

            AlgorithmModelTrackVo res = new AlgorithmModelTrackVo();

            BeanUtils.copyProperties(modelTrack, res);
            BeanUtils.copyProperties(algorithmConfigTrack, res);

            res.setAlgorithmId(algorithmConfigTrack.getId());
            res.setId(modelTrack.getId());

            return CommonResult.success(res, "success");
        } else if (entity.getName().startsWith(CAT)) {
            AlgorithmModelTrack modelTrack = algorithmModelTrackService.getById(entity.getAlgorithmModelId());
            AlgorithmConfigTrack algorithmConfigTrack = algorithmConfigTrackService
                    .getById(modelTrack.getAlgorithmId());

            AlgorithmModelTrackVo res = new AlgorithmModelTrackVo();

            BeanUtils.copyProperties(modelTrack, res);
            BeanUtils.copyProperties(algorithmConfigTrack, res);

            res.setAlgorithmId(algorithmConfigTrack.getId());
            res.setId(modelTrack.getId());

            return CommonResult.success(res, "success");
        } else {
            return CommonResult.fail("命名错误！");
        }
    }

    @Override
    public CommonResult getModelList(Long id) {
        TrackSequence entity = getById(id);
        String outPutPath = entity.getOutputPath();
        String modelPath = outPutPath;
        return getModelList_(modelPath, TaaisConfig.getProfile());
    }

    @Override
    public CommonResult previewPredictResult(Long id) {
        TrackSequence entity = getById(id);
        String inputPath = entity.getInputPath() + File.separator + "images";
        String outputPath = entity.getOutputPath() + File.separator + "gaze" + File.separator + "images";

        File outputPathDir = new File(outputPath);
        if (!outputPathDir.exists()) {
            outputPath = entity.getOutputPath() + File.separator + "images";
        }

        String urlPrefix = Constants.RESOURCE_PREFIX;
        return getCompareImage(urlPrefix, inputPath, outputPath);
    }

    @Override
    public CommonResult previewEvaluateResult(Long id) {
        return getStatisticsResult(id);
    }

    @Override
    public CommonResult getStatisticsResult(Long id) {
        TrackSequence entity = getById(id);

        String resultFilePath = entity.getOutputPath() + File.separator + BizConstant.RESULT_JSON_NAME;

        Path resultPath = Paths.get(resultFilePath);

        if (!new File(resultPath.toString()).exists()) {
            return CommonResult.fail("评估结果文件不存在！");
        }

        List<Dict> resultMap = parseJsonMapList(resultPath);
        if (ObjectUtil.isEmpty(resultMap)) {
            return CommonResult.fail("获取结果文件失败");
        }

        return CommonResult.success(resultMap);
    }

}
